import enum
import socket
from typing import Optional, Union

import pydantic
import pynng

from foreverbull_core.models.base import Base


class SocketType(enum.Enum):
    REQUESTER = pynng.Req0
    REPLIER = pynng.Rep0
    PUBLISHER = pynng.Pub0
    SUBSCRIBER = pynng.Sub0


class SocketConfig(Base):
    """_summary_

    Args:
        socket_type (SocketType): Union[SocketType, str] = SocketType.REPLIER
        host (str): str = socket.gethostbyname(socket.gethostname())
        port (int): int = 0
        listen (bool): bool = True
        recv_timeout (int): int = 500
        send_timeout (int): int = 500

    Returns:
        SocketConfig: _description_
    """

    socket_type: Union[SocketType, str] = SocketType.REPLIER
    host: str = socket.gethostbyname(socket.gethostname())
    port: int = 0
    listen: bool = True
    recv_timeout: int = 500
    send_timeout: int = 500

    @pydantic.validator("socket_type")
    def validate_socket_type(cls, v):
        if type(v) is str:
            return getattr(SocketType, v.upper())
        return v

    def dict(self, *args, **kwargs):
        return {
            "socket_type": self.socket_type.name,
            "host": self.host,
            "port": self.port,
            "listen": self.listen,
            "recv_timeout": self.recv_timeout,
            "send_timeout": self.send_timeout,
        }

    def dump(self):
        return self.dict()


class Request(Base):
    """_summary_

    Args:
        task (str): str
        data (dict, optional): Optional[dict] = None

    Returns:
        Request: request
    """

    task: str
    data: Optional[dict] = None


class Response(Base):
    """_summary_

    Args:
        task (str): str
        error (str, optional): Optional[str] = None
        data (dict, optional): Optional[dict] = None

    Returns:
        Response: response
    """

    task: str
    error: Optional[str] = None
    data: Optional[dict] = None



from foreverbull_core.models.socket import Request, Response, SocketConfig
from foreverbull_core.socket.nanomsg import NanomsgContextSocket, NanomsgSocket


class ContextClient:
    def __init__(self, context_socket: NanomsgContextSocket):
        """Context client is sub socket of SocketClient that will keep track of who sends the request
        to make sure respone will go to the same peer.

        Args:
            context_socket (NanomsgContextSocket): The context_socket to be used for communication,
            uses same Port/int as socket itself.
        """
        self._context_socket = context_socket

    def send(self, message: Response) -> None:
        """Sends a response back to the requester

        Args:
            message (Response): Response will be serialized to bytes before transmitting
        """
        self._context_socket.send(message.dump())

    def recv(self) -> Request:
        """Waits until incoming bytes has been received and load it into a Request Model

        Returns:
            Request: Request that has been received
        """
        data = self._context_socket.recv()
        return Request.load(data)

    def close(self) -> None:
        """Close the socket"""
        self._context_socket.close()


class SocketClient:
    def __init__(self, config: SocketConfig) -> None:
        """SocketClient provides a higher level connection for a socket intended to listen to incoming requests

        Args:
            config (SocketConfig): Configuration for how to socket should work, timeouts etc.
        """
        self.config = config
        self._socket = NanomsgSocket(config)

    def url(self) -> str:
        """Receive the connection information

        Returns:
            str: URL to the local connection
        """
        return self._socket.url()

    def send(self, message: Response) -> None:
        """Sends a response back to the requester

        Args:
            message (Response): Response will be serialized to bytes before transmitting
        """
        self._socket.send(message.dump())

    def recv(self) -> Request:
        """Waits until incoming bytes has been received and load it into a Request Model

        Returns:
            Request: Request that has been received
        """
        data = self._socket.recv()
        return Request.load(data)

    def close(self) -> None:
        """Close the socket"""
        self._socket.close()

    def new_context(self) -> ContextClient:
        """Returns a new context client for this socket that is intended when you might have several
        incoming connections at same time.
        Each context will know where the request came from and reply to the proper peer.

        Returns:
            ContextClient: A new conext client based in this Socket and its address.
        """
        return ContextClient(NanomsgContextSocket(self._socket.new_context()))



from pandas import DataFrame, read_sql_query
from sqlalchemy import create_engine, desc
from sqlalchemy.orm.session import sessionmaker

from foreverbull.data.stock_data import Base, Period, Position
from foreverbull_core.models.worker import Database as DatabaseConfiguration


class DateManager:
    def __init__(self, start, end):
        self.start = start
        self.end = end
        self.current = None


class Database:
    def __init__(self, execution_id: str, date_manager: DateManager, db_conf: DatabaseConfiguration = None):
        self.db_conf = db_conf
        if db_conf is None:
            self.uri = "sqlite:///:memory:"
        else:
            self.uri = (
                f"postgresql://{db_conf.user}:{db_conf.password}@{db_conf.netloc}:{db_conf.port}/{db_conf.dbname}"
            )
        self.execution_id = execution_id
        self.date_manager = date_manager
        self.engine = create_engine(self.uri)
        if self.db_conf is None:
            Base.metadata.create_all(self.engine)
        self.session_maker = sessionmaker(self.engine)

    def stock_data(self, symbol: str = None) -> DataFrame:
        if symbol:
            query = f"""Select symbol, time, high, low, open, close, volume
                        FROM ohlc WHERE time BETWEEN '{self.date_manager.start}'
                        AND '{self.date_manager.current}' AND symbol='{symbol}'"""
        else:
            query = f"""Select symbol, time, high, low, open, close, volume
                        FROM ohlc WHERE time BETWEEN '{self.date_manager.start}'
                        AND '{self.date_manager.current}'"""
        return read_sql_query(query, self.engine)

    def period(self) -> Period:
        with self.session_maker() as db_session:
            q = db_session.query(Period).filter_by(execution_id=self.execution_id)
            period = q.order_by(desc(Period.timestamp)).first()
        return period

    def get_position(self, symbol: str) -> Position:
        with self.session_maker() as db_session:
            query = db_session.query(Position)
            query = query.filter(Position.execution == self.execution_id)
            query = query.filter(symbol == Position.symbol)
            query = query.order_by(desc(Position.period))
            position = query.first()

        return position



import logging
import threading
from inspect import signature

from foreverbull.data import Database
from foreverbull.models import OHLC, Configuration, DefaultStrategy, Info, Parameter
from foreverbull.worker import WorkerPool
from foreverbull_core.socket.client import SocketClient, SocketConfig
from foreverbull_core.socket.exceptions import SocketClosed, SocketTimeout
from foreverbull_core.socket.router import MessageRouter


class Foreverbull(threading.Thread):
    _parameters = []
    _worker_routes = {}
    _default_strategy = None

    def __init__(
        self,
        socket_config: SocketConfig = None,
        worker_pool: WorkerPool = None,
        default_strategy: DefaultStrategy = None,
    ):
        self.socket_config = socket_config
        self.running = False
        self._worker_pool: WorkerPool = worker_pool
        self.logger = logging.getLogger(__name__)
        self._routes = MessageRouter()
        self._routes.add_route(self.info, "info")
        self._routes.add_route(self.stop, "stop")
        self._routes.add_route(self.configure, "configure", Configuration)
        self._routes.add_route(self.run_backtest, "run_backtest")
        threading.Thread.__init__(self)

    @staticmethod
    def _eval_param(type: str) -> str:
        if type == int:
            return "int"
        elif type == float:
            return "float"
        elif type == bool:
            return "bool"
        elif type == str:
            return "str"
        else:
            raise Exception("Unknown parameter type: {}".format(type))

    @staticmethod
    def add_route(func, msg_type):
        for key, value in signature(func).parameters.items():
            if value.annotation == Database or value.annotation == OHLC:
                continue
            default = None if value.default == value.empty else str(value.default)
            parameter = Parameter(key=key, default=default, type=Foreverbull._eval_param(value.annotation))
            Foreverbull._parameters.append(parameter)
        Foreverbull._worker_routes[msg_type] = func

    @staticmethod
    def on(msg_type, default_strategy: DefaultStrategy = None):
        def decorator(func):
            Foreverbull._default_strategy = default_strategy
            Foreverbull.add_route(func, msg_type)
            return func

        return decorator

    def info(self):
        return Info(
            type="worker", version="0.0.1", parameters=self._parameters, routes=list(self._worker_routes.keys())
        )

    def run(self):
        self.running = True
        self.logger.info("Starting instance")
        socket = SocketClient(self.socket_config)
        context_socket = None
        self.logger.info("Listening on {}:{}".format(self.socket_config.host, self.socket_config.port))
        while self.running:
            try:
                context_socket = socket.new_context()
                request = context_socket.recv()
                response = self._routes(request)
                context_socket.send(response)
                context_socket.close()
            except SocketTimeout:
                context_socket.close()
            except SocketClosed as exc:
                self.logger.exception(exc)
                self.logger.info("main socket closed, exiting")
        socket.close()
        self.logger.info("exiting")

    def configure(self, configuration: Configuration) -> None:
        self.logger.info("Configuring instance")
        self._worker_pool.configure(configuration)
        return None

    def run_backtest(self):
        self.logger.info("Running backtest")
        self._worker_pool.run_backtest()

    def stop(self):
        self.logger.info("Stopping instance")
        if self._worker_pool:
            self._worker_pool.stop()
        self.running = False



from datetime import datetime
from enum import IntEnum
from typing import Optional

from foreverbull_core.models.base import Base


class Asset(Base):
    symbol: str
    exchange: str
    name: str


class OHLC(Base):
    symbol: str
    open: float
    high: float
    low: float
    close: float
    volume: int
    time: datetime


class OrderStatus(IntEnum):
    OPEN = 0
    FILLED = 1
    CANCELLED = 2
    REJECTED = 3
    HELD = 4


class Order(Base):
    id: Optional[str]
    symbol: Optional[str]
    amount: Optional[int]
    filled: Optional[int]
    commission: Optional[int]
    limit_price: Optional[int]
    stop_price: Optional[int]
    created_at: Optional[datetime]
    status: Optional[OrderStatus]

    @classmethod
    def from_zipline(cls, order):
        return cls(
            id=order.id,
            symbol=order.sid.symbol,
            amount=order.amount,
            filled=order.filled,
            commission=order.commission,
            limit_price=order.limit,
            stop_price=order.stop,
            created_at=order.created,
            status=order.status,
        )


class Position(Base):
    symbol: str
    amount: int
    cost_basis: float
    period: datetime



import json
from typing import Union

from pydantic import BaseModel


class Base(BaseModel):
    """
    This is Base, here am i
    """

    @classmethod
    def load(cls, data: Union[dict, bytes]) -> object:
        """Loads a dynamic Dictionary or byte string containing dynamic Dictionary.
        Sets the inner key, values into the Pydantic Object

        Args:
            data (Union[dict, bytes]): Can either be a Dictionary or en encoded string of the Dictionary.

        Returns:
            object: Pydantic object that represents the Data.
        """
        if type(data) is dict:
            return cls(**data)
        loaded = json.loads(data.decode())
        return cls(**loaded)

    def dump(self) -> bytes:
        """_summary_

        Returns:
            bytes: _description_
        """
        return self.json().encode()

    def update_fields(self, object: dict):
        """_summary_

        Args:
            object (dict): _description_

        Returns:
            _type_: _description_
        """
        for key, value in object.items():
            if key in self.__fields__:
                setattr(self, key, value)
        return self



import logging
from collections import namedtuple
from typing import Callable

import pydantic

from foreverbull_core.models.socket import Request, Response

ROUTE = namedtuple("route", "func, route, model")


class TaskNotFoundError(Exception):
    pass


class TaskAlreadyExists(Exception):
    pass


class MessageRouter:
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._routes = {}

    def __call__(self, request: Request) -> Response:
        """Call the router with a Request, trying to find the saved route to call

        Args:
            request (Request): Request to send to a routed function

        Returns:
            Response: Response with possible data from the function called
        """
        if request.task not in self._routes:
            return Response(task=request.task, error=str(TaskNotFoundError("task not found")))
        route = self._routes[request.task]
        try:
            if route.model is None:
                data = route.func()
            else:
                model_data = route.model.load(request.data)
                data = route.func(model_data)
            return Response(task=request.task, data=data)
        except Exception as exc:
            self._logger.error(f"Error calling task: {request.task}")
            self._logger.error(exc, exc_info=True)
            return Response(task=request.task, error=repr(exc))

    def add_route(self, function: Callable, route: str, model: pydantic.BaseModel = None):
        """Add a route to the local _routes with a name<->function pair.

        Args:
            function (function): Function call shall be called on a route
            route (str): Name task which should call a specific function
            model (pydantic.BaseModel, optional): In case we shall include some data to the function being called.
            Defaults to None.

        Raises:
            TaskAlreadyExists: The task we are trying to add is already registed
        """
        if route in self._routes:
            raise TaskAlreadyExists(f"{route} already registered")
        new_route = ROUTE(function, route, model)
        self._routes[route] = new_route



from datetime import datetime

import requests

from foreverbull_core.models.socket import SocketConfig

from .exceptions import RequestError


class Service:
    def __init__(self, host: str, session: requests.session = None) -> None:
        """Initializes Service- endpoint client api

        Args:
            host (str): Host address to the Foreverbull backend server. IP:PORT Format
            session (requests.Session, optional): Use pre defined session instead of creating new. Defaults to None.
        """
        self.host = host
        if session is None:
            session = requests.Session()
        self.session: requests.Session = session

    def create(self, name: str, image: str) -> bool:
        rsp = self.session.put(
            f"http://{self.host}/api/v1/services",
            json={"name": name, "image": image},
        )
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""post call /services gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return True

    def get(self, service: str) -> dict:
        rsp = self.session.get(f"http://{self.host}/api/v1/services/{service}")
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""get call /services/{service} gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()

    def backtest_ingest(
        self, service: str, symbols: list, start: datetime, end: datetime, calendar: str, benchmark: str
    ) -> bool:
        rsp = self.session.put(
            f"http://{self.host}/api/v1/services/{service}/backtest/ingest",
            json={
                "symbols": symbols,
                "start_time": start.isoformat(),
                "end_time": end.isoformat(),
                "calendar": calendar,
                "benchmark": benchmark,
            },
        )
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""put call /services/{service}/backtest gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return True

    def update_instance(self, service: str, container_id: str, socket: SocketConfig, online: bool) -> bool:
        rsp = self.session.put(
            f"http://{self.host}/api/v1/services/{service}/instances/{container_id}",
            json={**socket.dict(), "online": online},
        )
        if not rsp.ok:
            code = rsp.status_code  # to mitigate next line too long
            raise RequestError(
                f"""get call /services/{service}/instances/{container_id} gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return True



import io

import minio
import pandas as pd


class Backtest:
    def __init__(self, client: minio.Minio) -> None:
        self.client = client

    def upload_backtest_result(self, backtest: str, result: pd.DataFrame) -> None:
        result.to_pickle("/tmp/result.pkl")
        self.client.fput_object("backtest-results", f"{backtest}", "/tmp/result.pkl")

    def download_backtest_results(self, backtest: str) -> pd.DataFrame:
        pickled_result = self.client.get_object("backtest-results", f"{backtest}").read()
        return pd.read_pickle(io.BytesIO(pickled_result))



import enum
from typing import Optional

import pydantic

from foreverbull_core.models.base import Base


class ServiceType(enum.Enum):
    WORKER = "worker"
    BACKTEST = "backtest"


class Service(Base):
    """_summary_

    Args:
        id (str, optional): Optional[str]
        name (str): str
        image (str): str
        type (ServiceType): ServiceType

    Returns:
        Service: _description_
    """

    id: Optional[str]
    name: str
    image: str
    type: ServiceType

    @pydantic.validator("type")
    def validate_type(cls, v):
        if type(v) is str:
            return getattr(ServiceType, v.upper())
        return v

    def dict(self, *args, **kwargs):
        return {
            "id": self.id,
            "name": self.name,
            "image": self.image,
            "type": self.type if type(self.type) is str else self.type.value,
        }


class Instance(Base):
    """_summary_

    Args:
        id (str): str
        service_id (str): str
        container_id (str, optional): Optional[str]
        host (str, optional): Optional[str]
        port (int, optional): Optional[int]

    Returns:
        Instance: instance
    """

    id: str
    service_id: str
    container_id: Optional[str]
    host: Optional[str]
    port: Optional[int]
    listen: Optional[bool]
    online: Optional[bool]


class RawConnection(Base):
    """_summary_

    Args:
        host (str): str
        port (int): int

    Returns:
        RawConnection: raw
    """

    host: str
    port: int



from foreverbull_core.http import HTTPClient
from foreverbull_core.models.socket import SocketConfig


class Broker:
    def __init__(self, broker_host: str, local_host: str) -> None:
        """Initializes a specific broker connected to Foreverbull

        Args:
            broker_host (str): Address of the server we are trying to connect
            local_host (str): Address of this localhost who is using this library
        """
        self._broker_host = broker_host
        self._local_host = local_host
        self.http = HTTPClient(self._broker_host)
        self.socket_config = SocketConfig(host=self._local_host, port=5555)



from datetime import datetime
from typing import List

import requests

from .exceptions import RequestError


class Finance:
    def __init__(self, host: str, session: requests.session = None) -> None:
        self.host = host
        if session is None:
            session = requests.Session()
        self.session: requests.Session = session

    def get_assets(self, symbols: List[str]) -> List[dict]:
        rsp = self.session.get(f"http://{self.host}/api/v1/finance/assets", params={"symbols": symbols})
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""get call /finance/assets gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()

    def create_assets(self, symbols: List[str]) -> List[dict]:
        rsp = self.session.put(f"http://{self.host}/api/v1/finance/assets", json={"symbols": symbols})
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""put call /finance/assets gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()

    def check_ohlc(self, symbols: List[str], start: datetime, end: datetime) -> bool:
        rsp = self.session.get(
            f"http://{self.host}/api/v1/finance/ohlc/check",
            params={"symbols": symbols, "startTime": start.isoformat(), "endTime": end.isoformat()},
        )
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""head call /finance/ohlc/check gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.ok

    def create_ohlc(self, symbols: List[str], start: datetime, end: datetime) -> List[dict]:
        rsp = self.session.put(
            f"http://{self.host}/api/v1/finance/ohlc",
            json={"symbols": symbols, "start_time": start.isoformat(), "end_time": end.isoformat()},
        )
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""put call /finance/ohlc gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()



import logging
import os
from functools import partial
from multiprocessing import Event, Process
from threading import Thread
from typing import List

import pynng

from foreverbull.data import Database, DateManager
from foreverbull.models import OHLC, Configuration, Parameter
from foreverbull.worker.exceptions import WorkerException
from foreverbull_core.models.socket import Request, Response
from foreverbull_core.socket.exceptions import SocketTimeout


class Worker:
    def __init__(self, survey_address: str, state_address: str, stop_event: Event, **routes):
        self.logger = logging.getLogger(__name__)
        self.logger.debug("setting up worker")
        self._survey_address = survey_address
        self._state_address = state_address
        self._stop_event = stop_event
        self._routes = routes
        self._func = routes.get("ohlc")
        self.parameters = {}
        self.logger.info("worker configured correctly")
        super(Worker, self).__init__()

    @staticmethod
    def _eval_param(type: str, val):
        if type == "int":
            return int(val)
        elif type == "float":
            return float(val)
        elif type == "bool":
            return bool(val)
        elif type == "str":
            return str(val)
        else:
            raise WorkerException("Unknown parameter type")

    def _setup_function(self, func, parameters: List[Parameter]):
        func = partial(func)
        for parameter in parameters:
            if parameter.value is None:
                parameter.value = parameter.default
            parameter.value = self._eval_param(parameter.type, parameter.value)
            func = partial(func, **{parameter.key: parameter.value})
        return func

    def _process_ohlc(self, ohlc: OHLC):
        self.logger.debug("Processing OHLC: %s, %s", ohlc.symbol, ohlc.time)
        self.date.current = ohlc.time
        try:
            return self._func(ohlc=ohlc, database=self.database)
        except KeyError:
            raise WorkerException("No route for ohlc")

    def configure(self, configuration: Configuration):
        self.configuration = configuration
        self.logger.info("configuring worker")
        self.socket = pynng.Rep0(dial=f"tcp://{self.configuration.socket.host}:{self.configuration.socket.port}")
        self.socket.recv_timeout = 500
        self.socket.send_timeout = 500
        if configuration.parameters:
            self._func = self._setup_function(self._func, configuration.parameters)
        self.date = DateManager(configuration.start_time, configuration.end_time)
        self.database = Database(configuration.execution, self.date, configuration.database)
        self.logger.info("worker configured correctly")

    def run(self):
        responder = pynng.Respondent0(dial=self._survey_address)
        responder.send_timeout = 10000
        responder.recv_timeout = 10000
        state = pynng.Pub0(dial=self._state_address)
        state.send(b"ready")
        self.logger.info("starting worker")
        while True:
            try:
                request = Request.load(responder.recv())
                self.logger.info("Received request")
                if request.task == "configure":
                    configuration = Configuration(**request.data)
                    self.configuration = configuration
                    self.configure(configuration)
                    responder.send(Response(task=request.task, error=None).dump())
                elif request.task == "run_backtest":
                    responder.send(Response(task=request.task, error=None).dump())
                    self.run_backtest()
                elif request.task == "stop":
                    responder.send(Response(task=request.task, error=None).dump())
                    responder.close()
                    state.close()
                    break
                else:
                    self.logger.info("Received unknown task")
                    responder.send(Response(task=request.task, error="Unknown task").dump())
            except pynng.exceptions.Timeout:
                pass
            except Exception as e:
                self.logger.exception(repr(e))
                responder.send(Response(task=request.task, error=repr(e)).dump())
                responder.close()
                state.close()
                raise WorkerException(repr(e))

    def run_backtest(self):
        while True:
            try:
                self.logger.debug("Getting context socket")
                context_socket = self.socket.new_context()
                request = Request.load(context_socket.recv())
                order = self._process_ohlc(OHLC(**request.data))
                self.logger.debug(f"Sending response {order}")
                context_socket.send(Response(task=request.task, data=order).dump())
                context_socket.close()
            except (SocketTimeout, pynng.exceptions.Timeout):
                context_socket.close()
            except Exception as e:
                self.logger.exception(repr(e))
                context_socket.send(Response(task=request.task, error=repr(e)).dump())
                context_socket.close()
            if self._stop_event.is_set():
                break
        self.socket.close()


class WorkerThread(Worker, Thread):
    pass


class WorkerProcess(Worker, Process):
    pass


class WorkerPool:
    def __init__(self, **routes):
        self.logger = logging.getLogger(__name__)
        self._workers = []
        self._routes = routes
        self._executors = 2
        self._stop_workers_event = Event()
        self._survey_address = "ipc:///tmp/worker_pool.ipc"
        self._state_address = "ipc:///tmp/worker_states.ipc"
        self.survey = pynng.Surveyor0(listen=self._survey_address)
        self.worker_states = pynng.Sub0(listen=self._state_address)
        self.worker_states.subscribe(b"")
        self.worker_states.recv_timeout = 10000
        self.survey.send_timeout = 30000
        self.survey.recv_timeout = 30000

    def setup(self):
        self.logger.info("starting workers")
        for i in range(self._executors):  # Hardcode to 4 workers for now
            self.logger.info("starting worker %s", i)
            if os.getenv("THREADED_EXECUTION"):
                worker = WorkerThread(
                    self._survey_address, self._state_address, self._stop_workers_event, **self._routes
                )
            else:
                worker = WorkerProcess(
                    self._survey_address, self._state_address, self._stop_workers_event, **self._routes
                )
            worker.start()
            self._workers.append(worker)
        responders = 0
        while True:
            try:
                self.worker_states.recv()
                responders += 1
                if responders == self._executors:
                    break
            except pynng.exceptions.Timeout:
                raise WorkerException("Workers did not respond in time")
        self.logger.info("workers started")

    def configure(self, configuration: Configuration):
        self.logger.info("configuring workers")
        self.survey.send(Request(task="configure", data=configuration).dump())
        responders = 0
        while True:
            try:
                self.survey.recv()
                responders += 1
                if responders == self._executors:
                    break
            except pynng.exceptions.Timeout:
                raise WorkerException("Workers did not respond in time")
        self.logger.info("workers configured")

    def run_backtest(self):
        self.logger.info("running backtest")
        self.survey.send(Request(task="run_backtest").dump())
        responders = 0
        while True:
            try:
                self.survey.recv()
                responders += 1
                if responders == self._executors:
                    break
            except pynng.exceptions.Timeout:
                raise WorkerException("Workers did not respond in time")
        self.logger.info("backtest running")

    def stop(self):
        if self._stop_workers_event.is_set():
            return
        self.logger.info("stopping workers")
        self._stop_workers_event.set()
        self.survey.send(Request(task="stop").dump())
        responders = 0
        while True:
            try:
                self.survey.recv()
                responders += 1
                if responders == self._executors:
                    break
            except pynng.exceptions.Timeout:
                raise WorkerException("Workers did not respond in time")
        self.logger.info("workers stopped")
        self.survey.close()
        self.worker_states.close()



from datetime import datetime, timezone
from typing import List, Optional

import numpy as np
from numpy import isnan

from foreverbull_core.models.base import Base
from foreverbull_core.models.socket import SocketConfig

# KEEP TILL HTTP IS FIXED


class Execution(Base):
    running: bool
    stage: str
    error: Optional[str]


class Session(Base):
    id: Optional[str]
    backtest_id: str
    worker_id: Optional[str]
    worker_count: Optional[int]
    run_automaticlly: bool = False
    execution: Optional[Execution]


# END KEEP


class Database(Base):
    user: str
    password: str
    netloc: str
    port: int
    dbname: str


class Sockets(Base):
    main: SocketConfig
    feed: SocketConfig
    broker: SocketConfig
    running: bool


class IngestConfig(Base):
    calendar: Optional[str]
    start_time: Optional[datetime]
    end_time: Optional[datetime]
    symbols: List[str]


class EngineConfig(Base):
    calendar: str
    start_time: datetime
    end_time: datetime
    benchmark: Optional[str] = None
    symbols: List[str]


class Period(Base):
    period: datetime
    shorts_count: Optional[int]
    pnl: Optional[int]
    long_value: Optional[int]
    short_value: Optional[int]
    long_exposure: Optional[int]
    starting_exposure: Optional[int]
    short_exposure: Optional[int]
    capital_used: Optional[int]
    gross_leverage: Optional[int]
    net_leverage: Optional[int]
    ending_exposure: Optional[int]
    starting_value: Optional[int]
    ending_value: Optional[int]
    starting_cash: Optional[int]
    ending_cash: Optional[int]
    returns: Optional[int]
    portfolio_value: Optional[int]
    longs_count: Optional[int]
    algo_volatility: Optional[int]
    sharpe: Optional[int]
    alpha: Optional[int]
    beta: Optional[int]
    sortino: Optional[int]
    max_drawdown: Optional[int]
    max_leverage: Optional[int]
    excess_return: Optional[int]
    treasury_period_return: Optional[int]
    benchmark_period_return: Optional[int]
    benchmark_volatility: Optional[int]
    algorithm_period_return: Optional[int]

    @classmethod
    def from_backtest(cls, period):
        return Period(
            period=period["period_open"].to_pydatetime().replace(tzinfo=timezone.utc),
            shorts_count=int(period["shorts_count"] * 100),
            pnl=int(period["pnl"] * 100),
            long_value=int(period["long_value"] * 100),
            short_value=int(period["short_value"] * 100),
            long_exposure=int(period["long_exposure"] * 100),
            starting_exposure=int(period["starting_exposure"] * 100),
            short_exposure=int(period["short_exposure"] * 100),
            capital_used=int(period["capital_used"] * 100),
            gross_leverage=int(period["gross_leverage"] * 100),
            net_leverage=int(period["net_leverage"] * 100),
            ending_exposure=int(period["ending_exposure"] * 100),
            starting_value=int(period["starting_value"] * 100),
            ending_value=int(period["ending_value"] * 100),
            starting_cash=int(period["starting_cash"] * 100),
            ending_cash=int(period["ending_cash"] * 100),
            returns=int(period["returns"] * 100),
            portfolio_value=int(period["portfolio_value"] * 100),
            longs_count=int(period["longs_count"] * 100),
            algo_volatility=None if isnan(period["algo_volatility"]) else int(period["algo_volatility"] * 100),
            sharpe=None if isnan(period["sharpe"]) else int(period["sharpe"] * 100),
            alpha=None if period["alpha"] is None or np.isnan(period["alpha"]) else int(period["alpha"] * 100),
            beta=None if period["beta"] is None or np.isnan(period["beta"]) else int(period["beta"] * 100),
            sortino=None if isnan(period["sortino"]) else int(period["sortino"] * 100),
            max_drawdown=None if isnan(period["max_drawdown"]) else int(period["max_drawdown"] * 100),
            max_leverage=None if isnan(period["max_leverage"]) else int(period["max_leverage"] * 100),
            excess_return=None if isnan(period["excess_return"]) else int(period["excess_return"] * 100),
            treasury_period_return=None
            if isnan(period["treasury_period_return"])
            else int(period["treasury_period_return"] * 100),
            benchmark_period_return=None
            if isnan(period["benchmark_period_return"])
            else int(period["benchmark_period_return"] * 100),
            benchmark_volatility=None
            if isnan(period["benchmark_volatility"])
            else int(period["benchmark_volatility"] * 100),
            algorithm_period_return=None
            if isnan(period["algorithm_period_return"])
            else int(period["algorithm_period_return"] * 100),
        )


class Result(Base):
    periods: List[Period]



from pynng import exceptions, nng

from foreverbull_core.models.socket import SocketConfig

from .exceptions import SocketClosed, SocketTimeout


class NanomsgContextSocket:
    def __init__(self, context_socket: nng.Context):
        """Provides a low level class for managing context sockets

        Args:
            context_socket (nng.Context): Context socket coming from a "normal" socket
        """
        self._context_socket = context_socket

    def send(self, data: bytes) -> None:
        """Send byte data over the socket to a peer

        Args:
            data (bytes): Bytes to send over the socket

        Raises:
            SocketClosed: In case the we are trying to send over a closed socket

        Returns:
            None:
        """
        try:
            return self._context_socket.send(data)
        except exceptions.Closed as exc:
            raise SocketClosed(exc)

    def recv(self) -> bytes:
        """Wait and receive byte data from a peer

        Raises:
            SocketTimeout: _description_
            SocketClosed: _description_

        Returns:
            bytes: Incoming byte data
        """
        try:
            return self._context_socket.recv()
        except exceptions.Timeout as exc:
            raise SocketTimeout(exc)
        except exceptions.Closed as exc:
            raise SocketClosed(exc)

    def close(self) -> None:
        """Close the socket"""
        return self._context_socket.close()


class NanomsgSocket:
    def __init__(self, config: SocketConfig):
        """Initializes a low level Nanomsg socket for transmitting data between peers

        Args:
            config (SocketConfig): Configuration about how the socket should work.
                                    In case Port is not set and we are a listener we will get a random port.
        """
        self._socket = None
        self._config = config
        self._socket = self._config.socket_type.value
        if self._config.listen:
            self._socket = self._config.socket_type.value(listen=f"tcp://{self._config.host}:{self._config.port}")
        else:
            self._socket = self._config.socket_type.value(dial=f"tcp://{self._config.host}:{self._config.port}")
        self._socket.recv_timeout = self._config.recv_timeout
        self._socket.send_timeout = self._config.send_timeout
        if self._config.listen and self._config.port == 0:
            # Pretty hacky way to find the port that OS randomly assigns when it's orginally set as 0
            self._config.port = int(self._socket.listeners[0].url.split(":")[-1])

    def url(self) -> str:
        """Returns the local address of the socket

        Returns:
            str: Local address in host:port format
        """
        if self._config.listen:
            return self._socket.listeners[0].url
        return self._socket.dialers[0].url

    def send(self, data: bytes) -> None:
        """Send byte data over the socket to a peer

        Args:
            data (bytes): Bytes to send over the socket

        Raises:
            SocketClosed: In case the we are trying to send over a closed socket

        Returns:
            None:
        """
        try:
            return self._socket.send(data)
        except exceptions.Closed as exc:
            raise SocketClosed(exc)

    def recv(self) -> bytes:
        """Wait and receive byte data from a peer

        Raises:
            SocketTimeout: _description_
            SocketClosed: _description_

        Returns:
            bytes: Incoming byte data
        """
        try:
            return self._socket.recv()
        except exceptions.Timeout as exc:
            raise SocketTimeout(exc)
        except exceptions.Closed as exc:
            raise SocketClosed(exc)

    def close(self) -> None:
        """Closes the socket"""
        return self._socket.close()

    def new_context(self) -> NanomsgContextSocket:
        """Returns a context socket used when we have multiple concurrent connections coming to us

        Returns:
            NanomsgContextSocket: Context socket based on this socket
        """
        return NanomsgContextSocket(self._socket.new_context())



import importlib
import logging
import os
import socket
import time
from multiprocessing import set_start_method

import click

import foreverbull_core.logger
from foreverbull import Foreverbull
from foreverbull.models import Configuration
from foreverbull.worker import WorkerPool
from foreverbull_core.broker import Broker
from foreverbull_core.models.socket import Request, SocketConfig, SocketType
from foreverbull_core.socket.client import SocketClient

foreverbull_core.logger.Logger()

log = logging.getLogger("Foreverbull")

broker = click.make_pass_decorator(Broker)


@click.group()
@click.argument("algo_file", type=str)
@click.pass_context
def base(ctx, algo_file: str):
    try:
        importlib.import_module(algo_file.replace("/", ".").split(".py")[0])
    except ModuleNotFoundError as e:
        logging.error(f"Could not import {algo_file} {e}")
        exit(1)

    broker_url = os.environ.get("BROKER_URL", "127.0.0.1:8080")
    local_host = os.environ.get("LOCAL_HOST", socket.gethostbyname(socket.gethostname()))
    ctx.obj = Broker(broker_url, local_host)


@base.command()
@broker
def start(broker: Broker):
    worker_pool = WorkerPool(**Foreverbull._worker_routes)
    worker_pool.setup()
    fb = Foreverbull(broker.socket_config, worker_pool)
    fb.start()

    while not broker.socket_config.port:
        time.sleep(0.1)

    try:
        broker.http.service.update_instance(
            os.environ.get("SERVICE_NAME"), socket.gethostname(), broker.socket_config, True
        )
        while fb.running:
            time.sleep(0.5)
    except KeyboardInterrupt:
        logging.info("Keyboard- Interrupt recieved exiting")
        broker.http.service.update_instance(
            os.environ.get("SERVICE_NAME"), socket.gethostname(), broker.socket_config, False
        )
    except Exception as e:
        logging.error(f"unable to call backend: {repr(e)}")

    fb.stop()
    exit(0)


@base.command()
@click.option("--strategy", type=str)
@broker
def run(broker: Broker, strategy: str):
    worker_pool = WorkerPool(**Foreverbull._worker_routes)
    worker_pool.setup()
    fb = Foreverbull(broker.socket_config, worker_pool)
    fb.start()

    while not broker.socket_config.port:
        time.sleep(0.1)

    if fb._default_strategy and strategy is None:
        try:
            stored_strategy = broker.http.strategy.get(fb._default_strategy.name)
            log.info(f"Using stored strategy: {stored_strategy}")
            if stored_strategy["backtest"] != fb._default_strategy.backtest_service:
                log.info("Stored backtest does not match default backtest. exiting...")
                fb.stop()
                exit(1)
            if stored_strategy["worker"] != fb._default_strategy.worker_service:
                log.info("Stored worker does not match default worker. exiting...")
                fb.stop()
                exit(1)
            strategy = stored_strategy["name"]
        except Exception:
            log.info("No stored strategy found. Creating one")
            pass

    # We never found a stored strategy, so we need to create one
    if fb._default_strategy and strategy is None:
        try:
            broker.http.finance.get_assets(fb._default_strategy.symbols)
            log.info("Assets already exist")
        except Exception:
            broker.http.finance.create_assets(fb._default_strategy.symbols)
            log.info("Assets created")

        try:
            broker.http.finance.check_ohlc(
                fb._default_strategy.symbols, fb._default_strategy.start, fb._default_strategy.end
            )
            log.info("OHLC already exists")
        except Exception:
            broker.http.finance.create_ohlc(
                fb._default_strategy.symbols, fb._default_strategy.start, fb._default_strategy.end
            )
            log.info("OHLC created")

        try:
            broker.http.service.create(fb._default_strategy.worker_service, fb._default_strategy.worker_service_image)
        except Exception as e:
            log.exception(e)
            log.info("exiting...")
            fb.stop()
            exit(1)

        log.info("Worker service created")

        try:
            broker.http.service.create(
                fb._default_strategy.backtest_service, fb._default_strategy.backtest_service_image
            )
        except Exception as e:
            log.exception(e)
            log.info("exiting...")
            fb.stop()
            exit(1)

        log.info("Backtest service created")
        for _ in range(20):
            rsp = broker.http.service.get(fb._default_strategy.backtest_service)

            if rsp["status"] == "ERROR":
                log.exception(rsp["error"])
                log.info("Exiting...")
                fb.stop()
                exit(1)

            if rsp["status"] == "READY":
                log.info("Backtest service ready")
                break
            time.sleep(1)

        broker.http.service.backtest_ingest(
            fb._default_strategy.backtest_service,
            fb._default_strategy.symbols,
            fb._default_strategy.start,
            fb._default_strategy.end,
            fb._default_strategy.calendar,
            fb._default_strategy.symbols[0],
        )
        log.info("Backtest service ingestion started")
        for _ in range(20):
            rsp = broker.http.service.get(fb._default_strategy.backtest_service)

            if rsp["status"] == "ERROR":
                log.exception(rsp["error"])
                log.info("Exiting...")
                fb.stop()
                exit(1)

            if rsp["status"] == "READY":
                log.info("Backtest service ready after ingestion")
                break
            time.sleep(1)
        broker.http.strategy.create(
            fb._default_strategy.name, fb._default_strategy.backtest_service, fb._default_strategy.worker_service
        )
        log.info("Strategy created")
        strategy = fb._default_strategy.name

    try:
        backtest = broker.http.backtest.run(strategy)
        while backtest["stage"] != "RUNNING":
            if backtest["error"]:
                log.error("backtest failed to start: ", backtest["error"])
                fb.stop()
                return
            time.sleep(1)
            backtest = broker.http.backtest.get(backtest["id"])
        socket_config = SocketConfig(
            host="127.0.0.1",
            port=27015,
            socket_type=SocketType.REQUESTER,
            listen=False,
            recv_timeout=10000,
            send_timeout=10000,
        )
        socket = SocketClient(socket_config)
        ctx = socket.new_context()
        ctx.send(Request(task="get_configuration"))
        rsp = ctx.recv()
        configuration = Configuration(**rsp.data)
        fb.configure(configuration)
        fb.run_backtest()
        ctx.send(Request(task="start"))
        rsp = ctx.recv()
        ctx.send(Request(task="stop"))
        rsp = ctx.recv()
    except KeyboardInterrupt:
        logging.info("Keyboard- Interrupt recieved exiting")
    except Exception as e:
        logging.error(f"unable to call backend: {repr(e)}")

    fb.stop()


if __name__ == "__main__":
    set_start_method("spawn")
    base()



import requests

from .exceptions import RequestError


class Strategy:
    def __init__(self, host: str, session: requests.session = None) -> None:
        self.host = host
        if session is None:
            session = requests.Session()
        self.session: requests.Session = session

    def create(self, name: str, backtest: str, worker: str):
        rsp = self.session.put(
            f"http://{self.host}/api/v1/strategies",
            json={"name": name, "backtest": backtest, "worker": worker},
        )
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""put call /strategies gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()

    def get(self, name: str):
        rsp = self.session.get(f"http://{self.host}/api/v1/strategies/{name}")
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""get call /strategies/{name} gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()



from foreverbull_core.http import backtest, finance, service, strategy


class HTTPClient:
    def __init__(self, host: str) -> None:
        """Initializes a HTTPClient which has the underlying http endpoints

        Args:
            host (str): Host address to the Foreverbull backend server. IP:PORT Format
        """
        self.backtest = backtest.Backtest(host)
        self.service = service.Service(host)
        self.finance = finance.Finance(host)
        self.strategy = strategy.Strategy(host)



from datetime import datetime, timezone
from typing import List, Optional

from pydantic import validator

from foreverbull_core.models.base import Base
from foreverbull_core.models.socket import SocketConfig
from foreverbull_core.models.worker import Database


class Parameter(Base):
    key: str
    default: str = None
    value: str = None
    type: str


class Configuration(Base):
    execution: str
    start_time: datetime
    end_time: datetime
    database: Optional[Database]
    parameters: Optional[List[Parameter]]
    socket: SocketConfig


class OHLC(Base):
    symbol: str
    open: float
    high: float
    low: float
    close: float
    volume: int
    time: datetime


class Info(Base):
    type: str
    version: str
    parameters: List[Parameter]
    routes: List[str]


class DefaultStrategy(Base):
    name: str
    symbols: List[str]
    start: datetime
    end: datetime
    calendar: str = "XNYS"
    backtest_service: str
    backtest_service_image: str
    worker_service: str
    worker_service_image: str
    parameters: Optional[List[Parameter]]

    @validator("start")
    def start_add_utc_tz_info(cls, v):
        return v.replace(tzinfo=timezone.utc)

    @validator("end")
    def end_add_utc_tz_info(cls, v):
        return v.replace(tzinfo=timezone.utc)



from sqlalchemy import Column, DateTime, Float, Integer, Numeric, String
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class OHLC(Base):
    __tablename__ = "ohlc"
    id = Column("id", Integer, primary_key=True)
    symbol = Column("symbol", String)
    open = Column("open", Numeric)
    high = Column("high", Numeric)
    low = Column("low", Numeric)
    close = Column("close", Numeric)
    volume = Column("volume", Integer)
    time = Column("time", DateTime)


class Position(Base):
    __tablename__ = "position"
    id = Column("id", Integer, primary_key=True)
    execution = Column("execution", String)
    symbol = Column("symbol", String)
    amount = Column("amount", Integer)
    cost_basis = Column("cost_basis", Float)
    period = Column("period", DateTime)


class Period(Base):
    __tablename__ = "backtest_period"
    id = Column("id", Integer, primary_key=True)
    execution_id = Column("execution", String)
    cash_flow = Column("cash_flow", Float)
    starting_cash = Column("starting_cash", Integer)
    portfolio_value = Column("portfolio_value", Float)
    pnl = Column("pnl", Float)
    returns = Column("_returns", Float)
    timestamp = Column("start_date", DateTime)
    positions_value = Column("positions_value", Float)
    positions_exposure = Column("positions_exposure", Float)



import requests

from foreverbull_core.http import RequestError


class Backtest:
    def __init__(self, host: str, session: requests.Session = None) -> None:
        """Initializes Backtest service client api

        Args:
            host (str): Host address to the Foreverbull backend server. IP:PORT Format
            session (requests.Session, optional): Use pre defined session instead of creating new. Defaults to None.
        """
        self.host = host
        if session is None:
            session = requests.Session()
        self.session = session

    def run(self, strategy: str) -> dict:
        rsp = self.session.put(
            f"http://{self.host}/api/v1/strategies/{strategy}/backtests",
            params={"execution_type": "manual"},
        )
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""put call /strategies/{strategy}/backtests gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()

    def get(self, backtest: str) -> dict:
        rsp = self.session.get(f"http://{self.host}/api/v1/strategies/backtests/{backtest}")
        if not rsp.ok:
            code = rsp.status_code
            raise RequestError(
                f"""get call /backtests/{backtest} gave bad return code: {code}
            Text: {rsp.text}"""
            )
        return rsp.json()



import minio

from .backtest import Backtest


class Storage:
    def __init__(self, address, access_key, secret_key, secure=False):
        self.client = minio.Minio(address, access_key=access_key, secret_key=secret_key, secure=secure)
        self.backtest = Backtest(self.client)

    def create_bucket(self, bucket_name):
        self.client.make_bucket(bucket_name)



from typing import NamedTuple

from foreverbull_core.models.base import Base


class Database(Base):
    """_summary_

    Args:
        user (str): str
        password (str): str
        netloc (str): str
        port (int): int
        dbname (str): str

    Returns:
        Database: database
    """

    user: str
    password: str
    netloc: str
    port: int
    dbname: str


class Run(NamedTuple):
    """_summary_

    Args:
        broker_url (str): str = ""
        local_host (str): str = ""

    Returns:
        Run: run
    """

    broker_url: str = ""
    local_host: str = ""


class BacktestRun(Run):
    """_summary_

    Args:
        service_id (str): str = ""
        instance_id (str): str =

    Returns:
        BacktestRun:
    """

    service_id: str = ""
    instance_id: str = ""


class TestRun(NamedTuple):
    """_summary_

    Args:
        NamedTuple (_type_): _description_
    """

    broker_url: str = ""
    local_host: str = ""
    backtest_id: str = ""



import logging
import logging.config


class Logger:
    def __init__(self):
        self.default_config = {
            "version": 1,
            "formatters": {"simple": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}},
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "DEBUG",
                    "formatter": "simple",
                    "stream": "ext://sys.stdout",
                }
            },
            "loggers": {"base": {"level": "DEBUG", "handlers": ["console"], "propagate": False}},
            "root": {"level": "DEBUG", "handlers": ["console"]},
        }
        logging.config.dictConfig(self.default_config)



from foreverbull.foreverbull import DefaultStrategy, Foreverbull
from foreverbull.models import Parameter
from foreverbull.worker import Worker

__all__ = [Foreverbull, Worker, DefaultStrategy, Parameter]
